module KoL.HttpLowlevel where

import Prelude hiding (read, catch)

import KoL.Util
import KoL.UtilTypes

import Logging

import qualified Data.ByteString
import qualified Data.ByteString.Char8
import Data.ByteString.Char8 (unpack, pack)
import Network.Socket
import Network.BSD
import Network.URI
import Network.HTTP
import Network.BufferType
import Network.Stream (ConnError(..), failWith, fmapE)
import Network.CGI (formEncode)
import System.IO
import Control.Applicative
import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Time.Clock
import Data.IORef

import Data.Maybe
import Data.Char
import Numeric (readHex)

-- import System.Random

doHTTPLOWLEVEL_DEBUG _ = return ()
-- doHTTPLOWLEVEL_DEBUG x = putStrLn $ "PROXY DEBUG: " ++ x

connGetBlock conn size = Data.ByteString.hGet conn size

connGetLine conn = do
	x <- Data.ByteString.hGetLine conn
-- 	putStrLn $ "connGetLine: " ++ (unpack x)
	return $ (unpack x) ++ "\n"

connPut conn d = Data.ByteString.hPut conn (pack $ d)

connFlush conn = hFlush conn

connGetContents conn = Data.ByteString.hGetContents conn

read_till_empty conn = do
	x <- connGetLine conn
	if (length x == 2) && (x == "\r\n")
		then return $ [x]
		else do
			xs <- read_till_empty conn
			return $ [x] ++ xs

getResponseHead conn = parseResponseHead <$> read_till_empty conn

switchResponse _ _ _ (Left e) _ = return $ Left e
switchResponse conn allow_retry bdy_sent (Right (cd, rn, hdrs)) rqst = do
-- 	putStrLn $ "DEBUG:switchResponse: " ++ (show (cd, rn, hdrs, rqst))
	x <- case matchResponse (rqMethod rqst) cd of
		Continue -> if bdy_sent
			then do {- keep waiting -}
				rsp <- getResponseHead conn
				switchResponse conn allow_retry bdy_sent rsp rqst
			else do {- Time to send the body -}
				connPut conn (rqBody rqst)
				connFlush conn
				rsp <- getResponseHead conn
				switchResponse conn allow_retry True rsp rqst
		Retry -> do {- Request with "Expect" header failed. Trouble is the request contains Expects other than "100-Continue" -}
			connPut conn (show rqst)
			connPut conn (rqBody rqst)
			connFlush conn
			rsp <- getResponseHead conn
			switchResponse conn False bdy_sent rsp rqst
		Done -> return $ Right $ Response cd rn hdrs (pack "")
		DieHorribly str -> return $ responseParseError "Invalid response:" str
		ExpectEntity -> do
			case lookupHeader HdrTransferEncoding hdrs of
				Just s -> case filter isAlphaNum $ map toLower s of
					"chunked" -> do
-- 						putStrLn $ "DEBUG:chunked"
						let parse_chunks = do
							l <- connGetLine conn
							let size = case readHex l of
								(s, _):_ -> s
								_ -> 0
							if size > 0
								then do
-- 									putStrLn $ "  get a chunk, size " ++ (show size) ++ "..."
									x <- connGetBlock conn size
									connGetLine conn
									xs <- parse_chunks
									return $ Data.ByteString.append x xs
								else do
									-- ... read footers, should store them ... --
									read_till_empty conn
									return (pack "")
						bdy <- parse_chunks
						let ftrs = []
						return $ Right $ Response cd rn (hdrs ++ ftrs) bdy
					_ -> return $ responseParseError "switchResponse" "Unknown Transfer-Encoding"
				Nothing -> case lookupHeader HdrContentLength hdrs of
					Just x -> case read_as x of
						Just size -> do
-- 							putStrLn $ "DEBUG:linear"
							bdy <- connGetBlock conn size
							return $ Right $ Response cd rn hdrs bdy
						_ -> return $ responseParseError "unrecognized content-length value" x
					Nothing -> do
						putStrLn $ "WARNING! No content-length header!"
						bdy <- connGetContents conn
						return $ Right $ Response cd rn hdrs bdy
-- 	putStrLn $ "DEBUG:/switchResponse: " ++ (show (cd,rn,hdrs,rqst))
	return x

mkreq useragent cookie absuri params =
		(absuri, normalizeRequest defaultNormalizeRequestOptions $ case params of
			Nothing -> Request { rqURI = absuri, rqMethod = GET, rqHeaders = cookiehdr, rqBody = "" }
			Just p -> let enc = formEncode p in Request { rqURI = absuri, rqMethod = POST, rqHeaders = cookiehdr ++ [mkHeader HdrContentType "application/x-www-form-urlencoded", mkHeader HdrContentLength (show $ length enc)], rqBody = enc })
	where
		cookiehdr = case cookie of
			Nothing -> [mkHeader HdrUserAgent useragent] -- ++ [mkHeader HdrConnection "Keep-Alive"]
			Just x -> [mkHeader HdrCookie x] ++ [mkHeader HdrUserAgent useragent] -- ++ [mkHeader HdrConnection "Keep-Alive"]

rewrite_headers hdrs = map (\(Header x y) -> (show x, y)) hdrs

doHTTPreq (absuri, rq) = do
	r <- Network.HTTP.simpleHTTP rq
	case r of
		Right resp -> return (absuri, rspBody resp, rewrite_headers $ rspHeaders resp, rspCode resp)
		x -> do
			putStrLn $ "simpleHTTP: " ++ (show x) ++ " while downloading " ++ (show rq)
			throwIO $ userError $ "simpleHTTP [" ++ (show x) ++ "]"

kolproxy_withVersion v hs = case v of
	[] -> hs
	(h:_)
		| h == httpVersion -> hs
		| otherwise -> (Header (HdrCustom "X-HTTP-Version") h) : hs

kolproxy_parseRequestHead x = case x of
	[] -> Left Network.Stream.ErrorClosed
	(com:hdrs) -> do
			(version, rqm, uri) <- requestCommand com
			hdrs' <- parseHeaders hdrs
			return (rqm, uri, kolproxy_withVersion version hdrs')
		where
			rqMethodMap = [
			   ("HEAD",    HEAD),
			   ("PUT",     PUT),
			   ("GET",     GET),
			   ("POST",    POST),
			   ("DELETE",  DELETE),
			   ("OPTIONS", OPTIONS),
			   ("TRACE",   TRACE),
			   ("CONNECT", CONNECT)]

			moduri uri = escapeURIString (\x -> x `notElem` "[]") uri

			requestCommand l = case (l, words l) of
				([], _) -> Network.Stream.failWith Network.Stream.ErrorClosed
				(_, (rqm:uri:version)) -> 
					case (parseURIReference $ moduri uri, lookup rqm rqMethodMap) of
						(Just u, Just r) -> return (version, r, u)
						(Just u, Nothing) -> return (version, Custom rqm, u)
						_ -> parse_err l
				_ -> parse_err l

			parse_err l = responseParseError "parseRequestHead" ("Request command line parse failure: " ++ l)

kolproxy_receiveHTTP conn = do
	let kolproxy_headerName x = map toLower (f $ f $ x)
		where f = reverse . dropWhile isSpace
			
	-- TODO: handle favicon.ico better?
	h <- Network.Stream.fmapE (\es -> kolproxy_parseRequestHead (map (Network.BufferType.buf_toStr Network.BufferType.bufferOps) es)) (readTillEmpty1 Network.BufferType.bufferOps (readLine conn))
-- 	putStrLn $ "getRequestHead: " ++ (show h)
	case h of
		Left x -> return $ Left x
		Right (rm, uri, hdrs) -> do
			Network.Stream.fmapE (\(ftrs,bdy) -> Right (Request uri rm (hdrs ++ ftrs) bdy)) $
				case lookupHeader HdrTransferEncoding hdrs of
					Nothing -> case lookupHeader HdrContentLength hdrs of
						Nothing -> return (Right ([], Network.BufferType.buf_empty Network.BufferType.bufferOps))
						Just x -> case reads x of
							((v,_):_) -> linearTransfer (readBlock conn) v
							_ -> return $ responseParseError "unrecognized Content-Length value" x
					Just hte -> if kolproxy_headerName hte == "chunked"
						then chunkedTransfer Network.BufferType.bufferOps (readLine conn) (readBlock conn)
						else uglyDeathTransfer "receiveHTTP"

mksocket = do
	proto <- Network.BSD.getProtocolNumber "tcp"
	bracketOnError (Network.Socket.socket Network.Socket.AF_INET Network.Socket.Stream proto) (Network.Socket.sClose) $ \s -> do
		Network.Socket.setSocketOption s Network.Socket.KeepAlive 1
		return s

kolproxy_openTCPConnection server = do
		bracketOnError (mksocket) (\s -> Network.Socket.sClose s) $ \s -> do
			-- TODO: use getAddrInfo and stuff? for ipv6??
			hostA <- getHostAddr hostname
			let a = Network.Socket.SockAddrInet (toEnum portnum) hostA
			Network.Socket.connect s a
-- 			r <- randomRIO (1, 100 :: Integer)
-- 			when (r < 10) $ throwIO $ userError $ "faked random connection error!"
			h <- Network.Socket.socketToHandle s ReadWriteMode
			return h
	where
		Just parsed = parseURI server
		Just auth = parseURIAuthority $ uriToAuthorityString parsed
		hostname = host auth
		portnum = fromMaybe 80 (port auth)
		getHostAddr h = do
			catchIO (Network.Socket.inet_addr hostname)    -- handles ascii IP numbers
				(\ _ -> do
					host <- getHostByName_safe hostname
					case Network.BSD.hostAddresses host of
						[]     -> fail ("openTCPConnection: no addresses in host entry for " ++ show h)
						(ha:_) -> return ha)
		getHostByName_safe h = 
			catchIO (Network.BSD.getHostByName h)
				(\ _ -> fail ("openTCPConnection: host lookup failure for " ++ show h))

mkconnthing server = do
	connmv <- newEmptyMVar
	let open_conn = do
		t <- getCurrentTime
		c <- kolproxy_openTCPConnection server
		rchan <- newChan
-- 		killref <- newIORef False
		req_counter <- newIORef 0
		let run = do
			(isok, (absuri, rq, mvdest, ref)) <- (readChan rchan) `catch` (\e -> do
				doHTTPLOWLEVEL_DEBUG $ "http readchan exception: " ++ (show (e :: SomeException))
				throwIO e)
			(what, was_last_request) <- case isok of
				Right (_requesting_it, last_request) -> log_time_interval_http_above ref 2000 ("HTTP reading: " ++ (show $ rqURI rq)) $ do
					what <- try $ (do
						rsp <- log_time_interval_http_above ref 500 "HTTP head" $ getResponseHead c -- fails to read from mafia because mafia terminates with LF instead of CRLF. Is this still true?
-- 						r <- randomRIO (1, 100 :: Integer)
-- 						when (r < 5) $ throwIO $ userError $ "faked random read error!"
						Right resp <- log_time_interval_http_above ref 1000 "HTTP body" $ switchResponse c True False rsp rq
	-- 					putStrLn $ "HTTP got " ++ (show absuri) -- ++ " - " ++ (unpack $ rspBody $ resp)
	-- 					putStrLn $ "HTTP got " ++ (show absuri) ++ show resp
	-- 					putStrLn $ "HTTP got " ++ (show absuri)
						return (absuri, rspBody resp, rewrite_headers $ rspHeaders resp, rspCode resp, resp)) `catch` (\e -> do
							doHTTPLOWLEVEL_DEBUG $ "http read exception: " ++ (show (e :: SomeException))
							throwIO e)
					return (what, last_request)
				Left err -> return (Left err, False)
			putMVar mvdest what
			going <- (modifyMVar connmv $ \(cf_stored, _t_stored, pending) -> do
				let kill_it = do
-- 					doHTTPLOWLEVEL_DEBUG $ "closed, making new connection"
					(cf, t, _p) <- open_conn
					let transfer n = when (n > 0) $ do
						(_, x) <- readChan rchan
						cf x
						transfer (n - 1)
					transfer (pending - 1)
					return ((cf, t, pending - 1), False)
				case (was_last_request, what) of
					(True, _) -> do
-- 						doHTTPLOWLEVEL_DEBUG $ "last request, refreshing connection"
						kill_it
					(_, Right (_, _, hdrs, _, _)) -> case lookup "Connection" hdrs of
						Just "close" -> do
-- 							doHTTPLOWLEVEL_DEBUG $ "got Just close, refreshing connection"
							kill_it
						_ -> do
							trefreshed <- getCurrentTime
							return ((cf_stored, trefreshed, pending - 1), True)
					_ -> do
						doHTTPLOWLEVEL_DEBUG $ "no Right hdrs for connection, refreshing"
						kill_it) `catch` (\e -> do
							doHTTPLOWLEVEL_DEBUG $ "http put exception for " ++ (uriPath absuri) ++ ": " ++ (show (e :: SomeException))
							throwIO e)
			when going run
		forkIO $ run `catch` (\e -> do
			doHTTPLOWLEVEL_DEBUG $ "http forked-run exception: " ++ (show (e :: SomeException))
			throwIO e)
		let cfunc (absuri, rq, mvdest, ref) = do
			isok <- log_time_interval_http ref ("HTTP asking: " ++ (show $ rqURI rq)) $ try $ do
				req_nr <- atomicModifyIORef req_counter (\x -> (x + 1, x + 1))
				let requesting_it = (req_nr <= 80)
				if requesting_it
					then do
-- 						doHTTPLOWLEVEL_DEBUG $ "HTTP asking for " ++ show (rqURI rq) ++ " (" ++ show req_nr ++ ")"
						connPut c (show rq)
						connPut c (rqBody rq)
-- 						r <- randomRIO (1, 100 :: Integer)
-- 						when (r < 5) $ throwIO $ userError $ "faked random write error!"
						connFlush c -- Maybe TODO???: only flush when done requesting???
					else do
-- 						putStrLn $ "waiting with request for " ++ show (rqURI rq) ++ " (" ++ show req_nr ++ ")"
						return ()
				return (requesting_it, req_nr == 80)
			writeChan rchan (isok, (absuri, rq, mvdest, ref))
		return (cfunc, t, 0)
	forkIO $ putMVar connmv =<< open_conn

	connchan <- newChan
	forkIO $ forever $ ((do
		x <- (readChan connchan) `catch` (\e -> do
			doHTTPLOWLEVEL_DEBUG $ "connchan read exception: " ++ (show (e :: SomeException))
			throwIO e)

		-- TODO: unify open_conn and transfer? should be 0 pending here
		modifyMVar_ connmv $ \(cf_stored, t_stored, pending) -> do
			tnow <- getCurrentTime
			(cf, t, p) <- if diffUTCTime tnow t_stored <= 60.0
				then do
-- 					doHTTPLOWLEVEL_DEBUG $ "not-stale, keeping connection | " ++ show (tnow, t_stored, diffUTCTime tnow t_stored)
					return (cf_stored, t_stored, pending)
				else do
-- 					doHTTPLOWLEVEL_DEBUG $ "stale, making new connection | " ++ show (tnow, t_stored, diffUTCTime tnow t_stored)
					open_conn
			cf x
			return (cf, t, p + 1)) `catch` (\e -> doHTTPLOWLEVEL_DEBUG $ "connchan error: " ++ (show (e :: SomeException))))
	return connchan :: IO ConnChanType


send_http_response h resp = do
	writeBlock h (Network.BufferType.buf_fromStr Network.BufferType.bufferOps $ show resp)
	writeBlock h (Data.ByteString.Char8.unpack $ rspBody resp)

end_http h = Network.HTTP.close h
