{-# LANGUAGE BangPatterns #-}

-- module LogParser (showLogs, compressLogs, parse_log, do_log_parsing, parse_line_data, parse_page_title, parse_line) where
module LogParser (showLogs, compressLogs) where

-- TODO: drill down to whether a page is useful ASAP, instead of having long lists around and parsing them
-- TODO: store parts of status separately??? means text gets non-identical

import Prelude hiding (read, catch)

import Data.Maybe
import Data.List
import KoL.Util
-- import KoL.UtilTypes
import System.IO
import Network.URI
import Text.Printf
import System.Directory
import Control.Applicative
import Text.JSON
-- import Data.Time.LocalTime
import Control.Monad
import Control.Concurrent
import Control.Exception
import Data.IORef

import qualified Codec.Compression.BZip
import qualified Data.ByteString.Lazy.Char8
import qualified Data.ByteString.Lazy.Internal
import qualified Database.SQLite3




data LogPageload = LogPageload {
	logidx :: !Integer,
	uri :: !URI,
	effuri :: !URI,
-- 	title :: Maybe String,
	turn_before :: !Integer,
	turn_after :: !Integer,
	ascensionday :: !Integer,
-- 	rolloverbefore :: Integer,
	zonename :: !String,
	zonecontainer :: !String,
-- 	allparams :: [(String, String)],
	stats_before :: !(Integer, Integer, Integer),
	stats_after :: !(Integer, Integer, Integer),
	trail_before :: !Bool,
	trail_after :: !Bool
} deriving (Show)

parse_page_title pagetext =
	case (pagetitle, monstername, advrestitle) of
		(_, [[x]], _) -> Just $ x
		(_, _, [[x]]) -> Just $ x
		([["Results:"]], _, _) -> Nothing
		([["Adventure Results:"]], _, _) -> Nothing
		([[x]], _, _) -> Just $ x
		_ -> Nothing
	where
		pagetitle = matchGroups "<tr><td style=\"color: white;\" align=center bgcolor=blue><b>([^<]+)</b></td></tr>" pagetext
		monstername = matchGroups ">You're fighting <span id='monname'>([^<]*)</span><" pagetext
		advrestitle = matchGroups "<td style=\"color: white;\" align=center bgcolor=blue><b>[a-zA-Z]* Results:</b></td></tr><tr><td style=\"padding: 5px; border: 1px solid blue;\"><center><table><tr><td><center><b>([^<]*)</b>" pagetext

format_number_range from to
	| (from == to) = "#"
	| (from == to - 1) = show from
	| (from < to) = (show $ from) ++ "-" ++ (show $ to - 1)

getstatus jscombinedstr = jso where
	Ok jscombobj = decodeStrict jscombinedstr :: Result (JSObject JSValue)
	Ok jso = valFromObj "status" jscombobj :: Result (JSObject JSValue)

get_value name jsobj = x where
	Just (JSString jsstr) = lookup name $ fromJSObject jsobj
	!x = fromJSString jsstr

get_value_int name jsobj = case valFromObj name jsobj of
	Ok (JSString s) -> Just jss where
		Just !jss = read_as (fromJSString s)
	Ok (JSRational _ r) -> Just rounded where
		!rounded = round r
	Ok JSNull -> Just 0 -- HACK for when rollover is broken and not set yet
	_ -> Nothing

isontrail jsobj = case valFromObj "effects" jsobj :: Result (JSObject JSValue) of
	Ok effectsobj -> isJust $ lookup "91635be2834f8a07c8ff9e3b47d2e43a" $ fromJSObject effectsobj
	_ -> False

build_thing lidx requri returi jsbefore jsafter __raw_row_cols = do
	if get_value_int "freedralph" jsbefore == Just 1
		then return Nothing
		else (do
			let Ok lastadvobjafter = valFromObj "lastadv" jsafter
			let getstats jsobj = (fromJust $ get_value_int "rawmuscle" jsobj, fromJust $ get_value_int "rawmysticality" jsobj, fromJust $ get_value_int "rawmoxie" jsobj)
			let !thing = LogPageload {
				logidx = fromJust $ read_as $ lidx,
				uri = requri,
				effuri = returi,
				turn_before = (fromJust $ get_value_int "turnsthisrun" jsbefore) + 1,
				turn_after = (fromJust $ get_value_int "turnsthisrun" jsafter) + 1,
				ascensionday = case get_value_int "daysthisrun" jsbefore of
					Just d -> d
					_ -> fromJust $ get_value_int "rollover" jsbefore,
				zonename = get_value "name" lastadvobjafter,
				zonecontainer = get_value "container" lastadvobjafter,
				stats_before = getstats jsbefore,
				stats_after = getstats jsafter,
				trail_before = isontrail jsbefore,
				trail_after = isontrail jsafter
			}
			let !__l = length $ show $ thing
			return $ Just thing) `catch` (\e -> do
					putStrLn $ "Error parsing row: " ++ show lidx ++ ": " ++ show (e :: SomeException)
					putStrLn $ "  contents: " ++ show __raw_row_cols
					return Nothing)

go_build_thing __raw_row_cols@[lidx, requestedurlstr, retrievedurlstr, statusbeforestr, statusafterstr] = do
	let Just requri = parseURIReference requestedurlstr
	let Just returi = parseURIReference retrievedurlstr
	let !jsbefore = getstatus statusbeforestr
	let !jsafter = getstatus statusafterstr
	build_thing lidx requri returi jsbefore jsafter __raw_row_cols

getresults s acc = do
	sr <- Database.SQLite3.step s
	case sr of
		Database.SQLite3.Row -> do
			[Database.SQLite3.SQLInteger lidx, Database.SQLite3.SQLText requestedurlstr, Database.SQLite3.SQLText retrievedurlstr, Database.SQLite3.SQLText statusbeforestr, Database.SQLite3.SQLText statusafterstr] <- Database.SQLite3.columns s
			newthing <- go_build_thing [show lidx, requestedurlstr, retrievedurlstr, statusbeforestr, statusafterstr]
			case newthing of
				Just x -> getresults s (x:acc)
				_ -> getresults s acc
		Database.SQLite3.Done -> return acc

scan_through_database filename = do
	log_db <- Database.SQLite3.open filename
	rsgood <- do
		s <- Database.SQLite3.prepare log_db "SELECT idx, requestedurl, retrievedurl, statusbefore, statusafter FROM pageloads;"
		r <- reverse <$> getresults s []
		Database.SQLite3.finalize s
		return r

	putStrLn $ filename ++ " length rsgood: " ++ (show $ length $ rsgood)

	let group_by_day x y = ascensionday x == ascensionday y
	let group_by_container x y = zonecontainer x == zonecontainer y
	let group_by_zone x y = zonename x == zonename y
	let group_by_turn x y = turn_before x == turn_before y
	let format_turn_range format xs =
			"<span class=\"turnrange inside\">" ++ (printf format tstr) ++ "</span>"
		where
			tstr = format_number_range from to
			from = turn_before (head xs)
			to = turn_after (last xs)
	let desc_statgain xs = printf "(%+d mus, %+d mys, %+d mox)" (samus - sbmus) (samys - sbmys) (samox - sbmox) :: String
		where
			(sbmus, sbmys, sbmox) = stats_before (head xs)
			(samus, samys, samox) = stats_after (last xs)
	let do_line parentheader (a, b, c) = return $ printf "%s %s" (format_turn_range c b) (concat parentheader ++ a) :: IO String
	let do_turn parentheader xs = do
		--mapM (\x -> do_line parentheader x) $ do_turn_group xs
		[[Just pagetext]] <- do_db_query log_db "SELECT pagetext FROM pageloads WHERE idx = ?;" [Just $ show $ logidx $ last xs]
		!y <- do_line parentheader $ (case parse_page_title pagetext of
			Just z -> z
			_ -> "??? " ++ (uriPath $ effuri $ last $ xs) ++ " ???", xs, "(%s)")
-- 		putStrLn y
		return $ if (turn_before $ head $ xs) == (turn_after $ last $ xs)
			then []
			else [y ++ "<br>\n"]

	zdivnamecounter <- newIORef (1 :: Integer)
	cdivnamecounter <- newIORef (1 :: Integer)
	ddivnamecounter <- newIORef (1 :: Integer)

	let do_level shortname do_sub namecounter group_method headliner parentheader xs = do
		let sn = shortname
		xid <- readIORef namecounter
		writeIORef namecounter (xid + 1)
		let subxs = groupBy group_method xs
		let desc = printf "%s %s" (format_turn_range "[%s]" xs) (concat parentheader ++ headliner xs) :: String
		subret <- mapM (\x -> do_sub [] x) subxs
		subret <- return $ concat subret
		case subret of
			[] -> do
				ret <- return $ []
				ret <- return $ (++) ret $ printf "<div id=\"%s%d\" class=\"%s\">\n" sn xid sn
				ret <- return $ (++) ret $ printf "%s %s<br>\n" desc (desc_statgain xs)
				ret <- return $ (++) ret $ "</div>\n"
				return [ret]
			[sr] -> return [sr]
			_ -> do
				ret <- return $ []
				ret <- return $ (++) ret $ printf "<div id=\"%s%d\" class=\"%s\">\n" sn xid sn
				ret <- return $ (++) ret $ printf "%s %s <a href=\"#\" onclick=\"return toggle(this, %sd%d);\" class=\"expander\">[+]</a><br>\n" desc (desc_statgain xs) sn xid
				ret <- return $ (++) ret $ printf "  <div id=\"%sd%d\" class=\"details\">\n" sn xid
				ret <- return $ ret ++ (concat subret)
				ret <- return $ (++) ret $ "  </div>\n"
				ret <- return $ (++) ret $ "</div>\n"
				return [ret]

	let do_zone = do_level "z" do_turn zdivnamecounter group_by_turn (\xs -> zonename $ head $ xs)
	let do_container = do_level "c" do_zone cdivnamecounter group_by_zone (\xs -> zonecontainer $ head $ xs)
	let do_day = do_level "d" do_container ddivnamecounter group_by_container (\xs -> printf "Day %d" $ ascensionday $ head $ xs :: String)

	all_lines <- mapM (\x -> do_day [] x) $ groupBy group_by_day rsgood

	let header_html = [
		"<html>",
		"<head>",
		"  <style>",
		"    body { font-family: monospace; }",
		"    .expander { color: green; }",
		"    .turnrange { color: blue; }",
		"    .z .details .turnrange { color: darkcyan; }",
		"    .inside { min-width: 12em; }",
		"    .details { margin-left: 20px; color: dimgray; display: none; }",
		"  </style>",
		"  <script type=\"text/javascript\">",
		"    function toggle(t, id) {",
		"      if (id.style.display == 'block') {",
		"        t.innerHTML = '[+]'",
		"        id.style.display = 'none'",
		"        return false",
		"      } else {",
		"        t.innerHTML = '[-]'",
		"        id.style.display = 'block'",
		"        return false",
		"      }",
		"    }",
		"  </script>",
		"</head>",
		"<body>"]

	let footer_html = [
		"</body>",
		"</html>"]

	let ret = header_html ++ (concat all_lines) ++ footer_html
-- 	let ret = []

	Database.SQLite3.close log_db
	return ret



-- import qualified Data.ByteString.Char8 as BSC8

-- import qualified Text.ParserCombinators.Parsec as P

-- import qualified Data.Attoparsec as AP
-- import qualified Data.Attoparsec.Combinator as APC
-- import qualified Data.Attoparsec.Char8 as APC8

-- import Control.Applicative

yielding_bs_eval Data.ByteString.Lazy.Internal.Empty = return Data.ByteString.Lazy.Internal.Empty
yielding_bs_eval (Data.ByteString.Lazy.Internal.Chunk x y) = do
	threadDelay 100000 -- delay to allow other things to run
	yy <- yielding_bs_eval y
	return $ Data.ByteString.Lazy.Internal.Chunk x yy

-- parsestr str = (x :: [(String, String)])
-- 	where Just x = read_as str

-- 	parsestr_ str = return x
-- 		where Just x = read_as str :: Maybe [(String, String)]




-- 	parseline' = P.sepBy parsecell' (P.char ',')
-- 	parsecell' = do
-- 		P.string "(\""
-- 		a <- P.many quotedChar'
-- 		P.string "\",\""
-- 		b <- P.many quotedChar'
-- 		P.string "\")"
-- 		return (a, b)
-- 	quotedChar' =
-- 		P.noneOf "\\\""
-- 		P.<|> P.try (P.string "\\\\" >> return '\\')
-- 		P.<|> P.try (P.string "\\\"" >> return '\"')

-- 	parsestr_' str = case P.parse parseline' "???" (tail $ init $ str) of
-- 		Right x -> return x
-- 		Left x -> throwIO $ userError $ "parsestr_' error: " ++ (show x)

-- 	parseline'' = (APC8.char '[') *> (parsecell'' `AP.sepBy` (APC8.char ',')) <* (APC8.char ']')
-- 	parsecell'' = do
-- 		APC8.char '('
-- 		APC8.char '"'
-- 		a <- quotedString''
-- 		APC8.char '"'
-- 		APC8.char ','
-- 		APC8.char '"'
-- 		b <- quotedString''
-- 		APC8.char '"'
-- 		APC8.char ')'
-- 		return (a, b)
-- 	-- 	return (BSC8.unpack a, BSC8.unpack b) -- TODO: unescape
-- 	quotedString'' = AP.many $
-- 		(APC8.char '\\' *> APC8.anyChar)
-- 		<|> (APC8.notChar '"')
-- 	-- quotedString'' = APC8.scan False (\x w -> case (x,w) of
-- 	-- 	(True, _) -> Just False
-- 	-- 	(False, '\\') -> Just True
-- 	-- 	(False, '"') -> Nothing
-- 	-- 	_ -> Just x)

-- 	parsestr_'' str = case AP.parseOnly parseline'' (BSC8.pack str) of
-- 		Right x -> return x
-- 		Left x -> throwIO $ userError $ "parsestr'' error: " ++ x

-- parse_page_title pagetext =
-- 	case (pagetitle, monstername, advrestitle) of
-- 		(_, [[x]], _) -> Just $ x
-- 		(_, _, [[x]]) -> Just $ x
-- 		([["Results:"]], _, _) -> Nothing
-- 		([["Adventure Results:"]], _, _) -> Nothing
-- 		([[x]], _, _) -> Just $ x
-- 		_ -> Nothing
-- 	where
-- 		pagetitle = matchGroups "<tr><td style=\"color: white;\" align=center bgcolor=blue><b>([^<]+)</b></td></tr>" pagetext
-- 		monstername = matchGroups ">You're fighting <span id='monname'>([^<]*)</span><" pagetext
-- 		advrestitle = matchGroups "<td style=\"color: white;\" align=center bgcolor=blue><b>[a-zA-Z]* Results:</b></td></tr><tr><td style=\"padding: 5px; border: 1px solid blue;\"><center><table><tr><td><center><b>([^<]*)</b>" pagetext



-- 	parse_line_data linenum pt params url effurl charstatus time = do
-- 		(turns, lastadv) <- case charstatus of
-- 			Just jscombined -> do
-- 	-- 			putStrLn x
-- 				let Ok jscomb = decodeStrict jscombined
-- 				let Ok jsobj = valFromObj "status" jscomb
-- 				let _get_value name = fromJSString jsstr
-- 					where Just (JSString jsstr) = lookup name $ fromJSObject jsobj
-- 				let get_value_int name = i
-- 					where i = case valFromObj name jsobj of
-- 						Ok i -> i
-- 						_ -> k
-- 							where
-- 								Just k = read_as j
-- 								Ok j = valFromObj name jsobj
-- 				let turns = get_value_int "turnsthisrun"
-- 				let lastadv = case valFromObj "lastadv" jsobj of
-- 					Ok i -> case (valFromObj "name" i, valFromObj "link" i) of
-- 						(Ok a, Ok b) -> Just (fromJSString a, fromJSString b)
-- 						_ -> throw $ InternalError $ "Error parsing last adventure"
-- 					_ -> throw $ InternalError $ "Error parsing last adventure"
-- 				return (Just turns, Just lastadv)
-- 			Nothing -> return (Nothing, Nothing)
-- 	-- 	putStrLn $ (show $ (turns, url, effurl, (\x -> take 20 x) <$> pt, lastadv))
-- 		let x = case (turns, url, params, effurl, pt, lastadv) of
-- 			(Just a, Just b, Just c, Just d, Just e, Just f) -> Just (a :: Integer, b :: URI, c :: Maybe [(String, String)], d :: URI, parse_page_title e, f :: Maybe (String, String))
-- 			_ -> Nothing
-- 		x <- if pt == Just "no|in a fight"
-- 			then return Nothing
-- 			else return x
-- 		putStrLn $ show (linenum, turns, url, time :: Maybe ZonedTime)
-- 		length (show x) `seq` return () -- force evaluation
-- 		return x

-- 	parse_line (linenum, line) = do
-- 		let dd = parsestr line
-- 		let pt = lookup "page text" dd >>= read_as
-- 		let params = lookup "parameters" dd >>= read_as
-- 		let url = lookup "requested url" dd >>= parseURIReference
-- 		let effurl = lookup "retrieved url" dd >>= parseURIReference
-- 		let charstatus = lookup "status" dd >>= read_as
-- 		let time = lookup "time" dd >>= read_as
-- 		parse_line_data linenum pt params url effurl charstatus time

-- 	backtrackToCoolStuff_ [] _init = Nothing
-- 	backtrackToCoolStuff_ ((turn, url, params, _effurl, title):xs) init
-- 		| ((uriPath url) `elem` ["/charpane.php", "/api.php", "/fight.php", "/choice.php", "/actionbar.php", "/skills.php", "/inv_equip.php", "/store.php"]) = backtrackToCoolStuff_ xs init
-- 	-- 	| ((uriPath url) `elem` ["/inv_use.php", "/multiuse.php"]) = Just (turn, url, params, title, init)
-- 		| otherwise = Just (turn, url, params, title, init)

-- 	backtrackToCoolStuff init = backtrackToCoolStuff_ init init

-- 	do_log_parsing parsed = do
-- 		let lastadvstuff = nub $ map (\(_a, _b, _c, _d, _e, f) -> case f of
-- 			Just (title, url) -> (fromJust $ (fromJust $ parseURIReference url) `relativeTo` (fromJust $ parseURIReference "/"), title)
-- 			_ -> throw $ InternalError $ "Error parsing last adventure") parsed
-- 		lastadvstuff <- return (lastadvstuff ++ [
-- 				(fromJust $ parseURIReference "/hiddencity.php", "The Hidden City")
-- 			])
-- 		putStrLn $ "lastadvstuff:"
-- 		mapM_ (\x -> putStrLn $ "  " ++ (show x)) lastadvstuff
-- 		let grouped = (reverse . (drop 1) . reverse) $ groupBy (\(turna, _, _, _, _) (turnb, _, _, _, _) -> turna == turnb) [(a, b, c, d, e) | (a, b, c, d, e, _f) <- parsed]
-- 		let counting = mapMaybe (backtrackToCoolStuff . reverse) grouped
-- 		let getsnarfblat url params = case (fromMaybe [] (decodeUrlParams url)) ++ params of
-- 			[] -> Nothing
-- 			xs -> case catMaybes [lookup "snarfblat" xs, lookup "adv" xs] of
-- 				z:_ -> fst <$> (listToMaybe $ reads z) :: Maybe Integer -- boss bat lair has snarfblat=34; bug
-- 				_ -> Nothing
-- 		let desceq url params (x, _y)
-- 			| (uriPath url == "/adventure.php") = and [uriPath x == uriPath url, isJust $ getsnarfblat url (fromMaybe [] params), getsnarfblat x [] == getsnarfblat url (fromMaybe [] params)]
-- 			| (uriPath url == "/main.php") = False
-- 			| otherwise = (uriPath url == uriPath x)
-- 		let descurl url params = case find (desceq url params) lastadvstuff of
-- 			Just (_, y) -> Right y
-- 			_ -> Left $ (show url) ++ " ???    " ++ (show params)
-- 		let collapsed = groupBy (\(_, urla, pa, _, _) (_, urlb, pb, _, _) -> descurl urla pa == descurl urlb pb) counting
-- 		let printed = map (\x -> case x of
-- 			[(a, b, c, _d, _e)] -> do
-- 				case descurl b c of
-- 					Right z -> printf "[%d] %s" (a+1) z
-- 					Left z -> do
-- 	-- 					let y = map (\x -> printf "  %s\n" (show x)) e
-- 	-- 					printf "[%d] %s -- ???\n%s\n" (a+1) z (concat y :: String)
-- 						printf "[%d] %s -- ???" (a+1) z
-- 			lots -> do
-- 				let (ha, hb, hc, _, _) = head lots
-- 				let (la, _lb, _lc, _, _) = last lots
-- 				case descurl hb hc of
-- 					Right z -> printf "[%d-%d] %s" (ha+1) (la+1) z
-- 					Left z -> do
-- 						printf "[%d-%d] %s -- ???" (ha+1) (la+1) z) collapsed
-- 	-- 	let descinit init = intercalate "\n" $ map (\(a,b,c,d) -> printf "  %s => %s | %s" (show b) (show c) (show d)) init
-- 	-- 	let printed = map (\(a,b,c,d) -> printf "[%d] %s | %s\n%s\n\n" (a+1) (descurl b) (show c) (descinit d)) counting
-- 		return (printed :: [String])

-- 	parse_log contents = do
-- 		parsed <- mapM parse_line $ zip ([1..] :: [Integer]) $ lines contents
-- 		parsed <- return $ catMaybes parsed
-- 		putStrLn $ "parsed length OK: " ++ (show $ length parsed)
-- 		do_log_parsing parsed

-- parse_line_xml line =
-- 		length (show x) `seq` x -- force evaluation
-- 	where
-- 		dd = parsestr line
-- 		pt = lookup "page text" dd >>= read_as
-- 		url = lookup "requested url" dd >>= parseURIReference
-- 		effurl = lookup "retrieved url" dd >>= parseURIReference
-- 		postparams = lookup "parameters" dd >>= read_as :: Maybe (Maybe [(String, String)])
-- 		charstatus = lookup "status" dd >>= read_as
-- 		state = lookup "state" dd >>= read_as :: Maybe (Maybe StateType)
-- 		(turnsplayed, jsdata) = case charstatus of
-- 			Just x -> (Just turns, Just x)
-- 				where
-- 					Ok jscomb = decodeStrict x :: Result (JSObject JSValue)
-- 					Ok jsobj = valFromObj "status" jscomb
-- 					get_value name = fromJSString jsstr
-- 						where Just (JSString jsstr) = lookup name $ fromJSObject jsobj
-- 					get_value_int name = i
-- 						where i = case valFromObj name jsobj of
-- 							Ok i -> i
-- 							otherwise -> k
-- 								where
-- 									Just k = read_as j
-- 									Ok j = valFromObj name jsobj
-- 					turns = get_value_int "turnsthisrun"
-- 			Nothing -> (Nothing, Nothing)
-- 		day = case state of
-- 			Just s -> unsafePerformIO $ do
-- 				stateref <- newIORef s
-- 				let ref = RefType { state = stateref }
-- 				daystr <- getState ref "session" "day"
-- 				case daystr of
-- 					Just ds -> case ((read_as ds) :: (Maybe Integer)) of
-- 						Just x -> return $ Just x
-- 						_ -> do
-- 							putStrLn $ "huh! session day is " ++ (show ds)
-- 							return Nothing
-- 					Nothing -> return Nothing
-- 			Nothing -> Nothing

-- 		x = case (day, turnsplayed, url, effurl, parse_page_title =<< pt, jsdata, postparams) of
-- 				(Just a, Just b, Just c, Just d, Just e, Just f, Just g) -> Just (a :: Integer, (b :: Integer, c :: URI, d :: URI, e, f :: String, g))
-- 				_ -> Nothing

-- 	printf "line %d [%s]: %s\n" (linenr :: Integer) (show turnsplayed) (show x)
-- 	length (show x) `seq` return () -- force evaluation
-- 	return x

-- 	parse_log_xml contents = do
-- 	-- 	let x = case (turns, url, params, effurl, pt, lastadv) of
-- 		let parsed = mapMaybe parse_line_xml (lines contents)
-- 		let grouped_day = groupBy (\(a,_) (b,_) -> a == b) parsed
-- 	-- 	putStrLn $ "parsed length: " ++ (show $ length parsed)
-- 		let handle_line day (turnsplayed, uri, effuri, t, jsdata, params)
-- 			| (uriPath effuri == "/fight.php") && (uriQuery effuri =~ "ireallymeanit") = do
-- 	-- 			printf "    %d-%d: %s -> %s = %s\n" day turnsplayed (show uri) (show effuri) (show pt)
-- 				let allparams = (concat $ maybeToList $ decodeUrlParams uri) ++ (concat $ maybeToList $ params)
-- 				let showuri = case allparams of
-- 					[] -> uri
-- 					_ -> x
-- 						where
-- 							testurienc = (uriPath uri) ++ "?" ++ (formEncode allparams)
-- 							testuri = escapeURIString (\x -> x `notElem` "[]") testurienc
-- 							Just x = parseURIReference testuri
-- 				printf "  [%d:%d] fight: %s [%s]\n" day turnsplayed t (show showuri)
-- 	-- 			return $ Just (day, "fight", t, showuri, jsdata)
-- 				let ret = Just ("fight", t, jsdata, showuri)
-- 				return (ret `seq` ret)
-- 			| otherwise = do
-- 	-- 			printf "    %d-%d: %s -> %s = %s [%s]\n" day turnsplayed (show uri) (show effuri) (show t) (show params)
-- 				return Nothing
-- 		xs <- mapM (\zzs -> do
-- 			let (day, _) = head zzs
-- 			z <- mapM (handle_line day) (map snd zzs)
-- 			return (day, catMaybes z)) grouped_day
-- 		let charname = "XXX"
-- 		let charclass = "XXX"
-- 		let ascnum = 12345
-- 		return (charname, charclass, ascnum, xs)


-- parseFile filename = do
-- 	contents <- readFile filename
-- 	parsed <- parse_log contents
-- 	return parsed

-- parseCompressedFile filename = do
-- 	compressed_contents <- Data.ByteString.Lazy.Char8.readFile filename
-- 	let contents = Data.ByteString.Lazy.Char8.unpack $ Codec.Compression.BZip.decompress compressed_contents
-- 	parsed <- parse_log contents
-- 	return parsed

-- showLogs fn = do
-- 	case fn of
-- 		Just filename -> do
-- 			bz2path <- getDirectoryPath "detailed log" (filename ++ "-detailed.txt.bz2")
-- 			bz2exists <- doesFileExist bz2path
-- 			info <- case bz2exists of
-- 				False -> do
-- 					path <- getDirectoryPath "detailed log" (filename ++ "-detailed.txt")
-- 					parseFile path
-- 				True -> parseCompressedFile bz2path
-- 			return $ "<html><pre>" ++ (intercalate "\n" info) ++ "</pre></html>"
-- 		_ -> do
-- 			basedir <- getBaseDirectory "detailed log"
-- 			filenames <- getDirectoryContents basedir
-- 			let describe x =
-- 				case matchGroups "^(.+)-([0-9]+)-detailed(.txt|.txt.bz2)$" x of
-- 					[[charname, ascnum, ".txt"]] -> Just (charname, fromJust $ read_as ascnum :: Integer, False)
-- 					[[charname, ascnum, ".txt.bz2"]] -> Just (charname, fromJust $ read_as ascnum, True)
-- 					_ -> Nothing
-- 			let ascs = sortBy (\(achar, aasc, _) (bchar, basc, _) -> compare (achar, (-aasc)) (bchar, (-basc))) $ mapMaybe describe filenames
-- 			let desc_asc_link (charname, ascnum, compressed) = case compressed of
-- 					True -> "<small>" ++ link ++ "</small>"
-- 					False -> link
-- 				where
-- 					link = "<a href=\"custom-logs?which=" ++ (charname ++ "-" ++ (show ascnum)) ++ "\">" ++ (show ascnum) ++ "</a>"
-- 			let group_to_row x = case x of
-- 				[] -> ""
-- 				(achar, _, _):_ -> "<tr><td>" ++ achar ++ "</td><td>" ++ (intercalate ", " (map desc_asc_link x)) ++ "</td></tr>\n"
-- 			let groupem xs =
-- 					map group_to_row grouped
-- 				where
-- 					grouped = groupBy (\(achar, _, _) (bchar, _, _) -> achar == bchar) xs
-- 			let rowtexts = groupem ascs
-- 			let pt = "<html><table>\n" ++ (concat rowtexts) ++ "</table></html>"
-- 			return pt

showLogs fn = do
	case fn of
		Just basefilename -> do
			bz2path <- getDirectoryPath "sqlite3 log" (basefilename ++ ".ascension-log.sqlite3.bz2")
			bz2exists <- doesFileExist bz2path
			lines <- case bz2exists of
				False -> do
					path <- getDirectoryPath "sqlite3 log" (basefilename ++ ".ascension-log.sqlite3")
					putStrLn $ "processing " ++ path
					withBinaryFile "temp-logparse-file.sqlite3" WriteMode $ \handle_out -> do
						withBinaryFile path ReadMode $ \handle_in -> do
							Data.ByteString.Lazy.Char8.hGetContents handle_in >>= Data.ByteString.Lazy.Char8.hPut handle_out
					putStrLn $ "  parsing log"
					scan_through_database "temp-logparse-file.sqlite3"
				True -> do
					putStrLn $ "processing compressed " ++ bz2path
					withBinaryFile "temp-logparse-file.sqlite3" WriteMode $ \handle_out -> do
						withBinaryFile bz2path ReadMode $ \handle_in -> do
							Codec.Compression.BZip.decompress <$> Data.ByteString.Lazy.Char8.hGetContents handle_in >>= Data.ByteString.Lazy.Char8.hPut handle_out
					putStrLn $ "  parsing log"
					scan_through_database "temp-logparse-file.sqlite3"
			putStrLn $ "log parsing done!"
			return $ intercalate "\n" lines
		_ -> do
			basedir <- getBaseDirectory "sqlite3 log"
			filenames <- getDirectoryContents basedir
			let describe x =
				case matchGroups "^(.+)-([0-9]+)(.ascension-log.sqlite3|.ascension-log.sqlite3.bz2)$" x of
					[[charname, ascnum, ".ascension-log.sqlite3"]] -> Just (charname, fromJust $ read_as ascnum :: Integer, False)
					[[charname, ascnum, ".ascension-log.sqlite3.bz2"]] -> Just (charname, fromJust $ read_as ascnum, True)
					_ -> Nothing
			let ascs = sortBy (\(achar, aasc, _) (bchar, basc, _) -> compare (achar, (-aasc)) (bchar, (-basc))) $ mapMaybe describe filenames
			let desc_asc_link (charname, ascnum, compressed) = case compressed of
					True -> "<small>" ++ link ++ "</small>"
					False -> link
				where
					link = "<a href=\"custom-logs?which=" ++ (charname ++ "-" ++ (show ascnum)) ++ "\">" ++ (show ascnum) ++ "</a>"
			let group_to_row x = case x of
				[] -> ""
				(achar, _, _):_ -> "<tr><td>" ++ achar ++ "</td><td>" ++ (intercalate ", " (map desc_asc_link x)) ++ "</td></tr>\n"
			let groupem xs =
					map group_to_row grouped
				where
					grouped = groupBy (\(achar, _, _) (bchar, _, _) -> achar == bchar) xs
			let rowtexts = groupem ascs
			let pt = "<html><body>Click on a log to parse it. Parsing a log takes several minutes, and the format and features are still being developed.<table>\n" ++ (concat rowtexts) ++ "</table></body></html>"
			return pt

compressFile path = do
	let bz2path = path ++ ".bz2"
	bz2exists <- doesFileExist bz2path
	if bz2exists
		then do
			putStrLn $ "Error: " ++ (show bz2path) ++ " already exists."
		else do
			putStrLn $ "compressing " ++ path
			contents <- Data.ByteString.Lazy.Char8.readFile path
			compressed_contents <- yielding_bs_eval $ Codec.Compression.BZip.compress contents
			h <- openFile bz2path WriteMode
			Data.ByteString.Lazy.Char8.hPut h compressed_contents
			hClose h
			putStrLn $ "compressed " ++ path

			test_uncompressed <- Data.ByteString.Lazy.Char8.readFile path
			test_compressed <- Data.ByteString.Lazy.Char8.readFile bz2path
			decompressed <- yielding_bs_eval test_compressed >>= return . Codec.Compression.BZip.decompress
			if test_uncompressed == decompressed
				then do
					putStrLn $ "compression ok, removing file " ++ path
					removeFile path
				else do
					putStrLn $ "Error: compressed file is different from the original."
			return ()

compressLogs charname asc = do
	let check_file x = do
		path <- getDirectoryPath "detailed log" (charname ++ "-" ++ (show x) ++ "-detailed.txt")
		exists <- doesFileExist path
-- 		putStrLn $ "check " ++ show (path, exists)
		when exists $ compressFile path
		path_2 <- getDirectoryPath "sqlite3 log" (charname ++ "-" ++ (show x) ++ ".ascension-log.sqlite3")
		exists_2 <- doesFileExist path_2
-- 		putStrLn $ "check2 " ++ show (path_2, exists_2)
		when exists_2 $ compressFile path_2
		
-- 	putStrLn $ "compressing old logs... " ++ charname ++ ": " ++ (show [1..(asc-1)])
	mapM_ check_file [1..(asc-1)]
